/* 
 * Weston standard cell constant temperature bath controller
 * File:   main.c
 * Author: Hirofumi Momose (hirofumi[at]mmomose.com)
 *
 * Created on 2018/08/26, 7:56
 * $Rev: 25 $
 */

/******************************************************************************/
/* INCLUDES                                                                   */
/******************************************************************************/
#include <xc.h>
#include <stdio.h>
#include <ctype.h>
#include "usb/inc/usb.h"
#include "usb/inc/usb_device.h"
#include "usb/inc/usb_device_cdc.h"
#include "usb/inc/usb_config.h"

/******************************************************************************/
/* CONFIGURATION Bits                                                         */
/******************************************************************************/
/* Register: CONFIG1L @ 0x300000 */
#pragma config LS48MHZ = SYS48X8 /* Low Speed USB mode with 48 MHz system clock */
// SYS48X8 System clock at 48 MHz, USB clock divider is set to 8
// SYS24X4 System clock at 24 MHz, USB clock divider is set to 4 (default)
#pragma config CPUDIV = CLKDIV3 /* CPU System Clock Postscaler */
// CLKDIV6  CPU uses system clock divided by 6
// CLKDIV3  CPU uses system clock divided by 3
// CLKDIV2  CPU uses system clock divided by 2
// NOCLKDIV CPU uses system clock (no divide) (default)
#pragma config CFGPLLEN = OFF /* PLL Enable Configuration bit */
// ON   PLL Enabled
// OFF  PLL Disabled (firmware controlled) (default)
#pragma config PLLSEL = PLL3X /* PLL Selection */
// PLL4X  4x clock multiplier (default)
// PLL3X  3x clock multiplier

/* Register: CONFIG1H @ 0x300001 */
#pragma config IESO = OFF /* Internal/External Oscillator Switchover */
// ON   Oscillator Switchover mode enabled
// OFF  Oscillator Switchover mode disabled (default)
#pragma config FCMEN = OFF /* Fail-Safe Clock Monitor */
// ON   Fail-Safe Clock Monitor enabled
// OFF  Fail-Safe Clock Monitor disabled (default)
#pragma config PCLKEN = OFF /* Primary Oscillator Shutdown */
// ON   Primary oscillator enabled (default)
// OFF  Primary oscillator shutdown firmware controlled
#pragma config FOSC = INTOSCIO /* Oscillator Selection */
// ECHCLKO    EC oscillator, high power 16MHz to 48MHz, clock output on OSC2
// HSH        HS oscillator, high power 16MHz to 25MHz
// ECLIO      EC oscillator, low power <4MHz
// ECLCLKO    EC oscillator, low power <4MHz, clock output on OSC2
// RCCLKO     External RC oscillator, clock output on OSC2
// ECMCLKO    EC oscillator, medium power 4MHz to 16MHz, clock output on OSC2
// INTOSCIO   Internal oscillator
// LP         LP oscillator
// ECMIO      EC oscillator, medium power 4MHz to 16MHz
// HSM        HS oscillator, medium power 4MHz to 16MHz
// XT         XT oscillator
// ECHIO      EC oscillator, high power 16MHz to 48MHz (default)
// RCIO       External RC oscillator
// INTOSCCLKO Internal oscillator, clock output on OSC2

/* Register: CONFIG2L @ 0x300002 */
#pragma config nLPBOR = ON /* Low-Power Brown-out Reset */
// OFF  Low-Power Brown-out Reset disabled (default)
// ON   Low-Power Brown-out Reset enabled
#pragma config BORV = 220 /* Brown-out Reset Voltage */
// 220  BOR set to 2.2V nominal
// 285  BOR set to 2.85V nominal
// 250  BOR set to 2.5V nominal
// 190  BOR set to 1.9V nominal (default)
#pragma config BOREN = SBORDIS /* Brown-out Reset Enable */
// NOSLP    BOR enabled in hardware, disabled in Sleep mode (SBOREN is ignored)
// ON       BOR controlled by firmware (SBOREN is enabled)
// OFF      BOR disabled in hardware (SBOREN is ignored)
// SBORDIS  BOR enabled in hardware (SBOREN is ignored) (default)
#pragma config nPWRTEN = OFF /* Power-up Timer Enable */
// OFF  Power up timer disabled (defalut)
// ON   Power up timer enabled

/* Register: CONFIG2H @ 0x300003 */
#pragma config WDTPS = 256 /* Watchdog Timer Postscaler */
// 32768    1:32768     131s (default)
// 16384    1:16384      65s
// 8192     1:8192       32s
// 4096     1:4096       16s
// 2048     1:2048        8s
// 1024     1:1024        4s
// 512      1:512         2s
// 256      1:256         1s
// 128      1:128      512ms
// 64       1:64       256ms
// 32       1:32       128ms
// 16       1:16        64ms
// 8        1:8         32ms
// 4        1:4         16ms
// 2        1:2          8ms
// 1        1:1          4ms
#pragma config WDTEN = ON /* Watchdog Timer Enable bits */
// SWON     WDT controlled by firmware (SWDTEN enabled)
// NOSLP    WDT enabled in hardware, disabled in Sleep mode (SWDTEN ignored)
// ON       WDT enabled in hardware (SWDTEN ignored) (defalut)
// OFF      WDT disabled in hardware (SWDTEN ignored)

/* Register: CONFIG3H @ 0x300005 */
#pragma config MCLRE = ON /* Master Clear Reset Pin Enable */
// ON   MCLR pin enabled; RE3 input disabled (defalut)
// OFF  RE3 input pin enabled; external MCLR disabled
#pragma config SDOMX = RB3 /* SDO Output MUX bit */
// RB3  SDO function is on RB3 (default)
// RC7  SDO function is on RC7
#pragma config T3CMX = RC0 /* Timer3 Clock Input MUX bit */
// RC0  T3CKI function is on RC0 (default)
// RB5  T3CKI function is on RB5
#pragma config PBADEN = ON /* PORTB A/D Enable bit */
// ON   PORTB<5:0> pins are configured as analog input channels on Reset (default)
// OFF  PORTB<5:0> pins are configured as digital I/O on Reset
#pragma config CCP2MX = RC1 /* CCP2 MUX bit */
// RC1  CCP2 input/output is multiplexed with RC1 (default)
// RB3  CCP2 input/output is multiplexed with RB3

/* Register: CONFIG4L @ 0x300006 */
#pragma config DEBUG = OFF /* Background Debugger Enable bit */
// OFF  Background debugger disabled, RB6 and RB7 configured as general purpose I/O pins (default)
// ON   Background debugger enabled, RB6 and RB7 are dedicated to In-Circuit Debug
#pragma config XINST = OFF /* Extended Instruction Set Enable bit */
// ON   Instruction set extension and Indexed Addressing mode enabled
// OFF  Instruction set extension and Indexed Addressing mode disabled (default)
#pragma config ICPRT = ON /* Dedicated In-Circuit Debug/Programming Port Enable */
// ON   ICPORT enabled (default)
// OFF  ICPORT disabled
#pragma config LVP = ON /* Single-Supply ICSP Enable bit */
// ON   Single-Supply ICSP enabled if MCLRE is also 1 (default)
// OFF  Single-Supply ICSP disabled
#pragma config STVREN = ON /* Stack Full/Underflow Reset */
// ON   Stack full/underflow will cause Reset (default)
// OFF  Stack full/underflow will not cause Reset

/******************************************************************************/
/* Data types                                                                 */
/******************************************************************************/
// Transfer function state buffer
struct tfbuf_s {
    float in_z1;   // Previous input (1/z)
    float out_z1;  // Previous output  (1/z)
};

// Config parameter
union conf_u {
    struct conf_s{
        float tset;             // Set Temperature
        float pgain;            // P gain
        float igain;            // I gain
        float dgain;            // D gain
        unsigned char src;      // Source (0=avg, 1=sensor1, 2=sensor2)
        unsigned char tint;     // Talker interval (0=disable)
        unsigned char flags;    // flags
    } st;
    unsigned char data[sizeof(struct conf_s)];
};

// Dialog state
enum dlg_e {
    DLG_NONE,
    DLG_MENU,
    DLG_TSET,
    DLG_PGAIN,
    DLG_IGAIN,
    DLG_DGAIN,
    DLG_SRC,
    DLG_TINT,
    DLG_FWUP
};

/******************************************************************************/
/* Global variables                                                           */
/******************************************************************************/
struct tfbuf_s tfbuf[4];
union conf_u conf;
enum dlg_e dlg = DLG_NONE;
unsigned char xtsk_100ms = 0;
unsigned char disp_seq = 0;
unsigned long tick = 0;

/******************************************************************************/
/* Function prototypes                                                        */
/******************************************************************************/
// Main
void main();
void main_halt();

// Interrupt handler
void interrupt high_priority isr_high();
void interrupt low_priority isr_low();

// Talker
void talk(unsigned long tick, float temp0, float temp1, float temp, float duty, float vcc);

// Parameter configuration
void conf_read();
void conf_write();
void conf_show();
void conf_dialog();

// LPF, PID-controller 
void tfbuf_init(float in, struct tfbuf_s *pbuf);
float lpf(float in, struct tfbuf_s *pbuf);
float pid_control(float obj, float curr, float pgain, float igain, float dgain, struct tfbuf_s *ibuf, struct tfbuf_s *dbuf);

// Temperature sensor
void temp_init(unsigned char ch);
float temp_read(unsigned char ch);

// PWM duty control 
void pwm_setduty(float duty);

// A/D conversion (VCC measurement)
float adc_getvcc(void);

// LED Display
void led_blank(void);
void led_digit(unsigned char digit);
void disp_debug();
void disp_ul(unsigned long ul);
void disp_temp(float temp);
void disp_err(unsigned char err);

// EEPROM
unsigned char eep_read(unsigned char addr);
void eep_write(unsigned char addr, unsigned char data);

// FW Loader invocation
void invoke_fwloader(void);

// I2C
#define i2c_start()     do {SSP1CON2bits.SEN = 1; while(SSP1CON2bits.SEN) continue;} while(0)
#define i2c_restart()   do {SSP1CON2bits.RSEN = 1; while (SSP1CON2bits.RSEN) continue;} while(0)
#define i2c_stop()      do {SSP1CON2bits.PEN = 1; while (SSP1CON2bits.PEN) continue; } while(0)
#define i2c_send(c)     do {PIR1bits.SSPIF = 0; SSP1BUF = c; while (!PIR1bits.SSPIF) continue; } while(0)
#define i2c_isacked()   (!SSP1CON2bits.ACKSTAT)
#define i2c_sendack()   do {SSP1CON2bits.ACKDT = 0; SSP1CON2bits.ACKEN = 1; while (SSP1CON2bits.ACKEN) continue; } while(0)
#define i2c_sendnack()  do {SSP1CON2bits.ACKDT = 1; SSP1CON2bits.ACKEN = 1; while (SSP1CON2bits.ACKEN) continue; } while(0)
unsigned char i2c_recv();

/******************************************************************************/
/* Macro                                                                      */
/******************************************************************************/
#define SEG_DP          0x20
#define SEG_DASH        0x10
#define SEG_BLANK       0x11

#define SYS_FREQ        16000000L       // 16MHz
#define _XTAL_FREQ      SYS_FREQ
#define FCY             SYS_FREQ/4
#define TMR0CLK         (FCY/16)        // 1MHz
#define MSEC2TMR0(x)    -(TMR0CLK/1000*(x))
#define TMR1CLK         (FCY/4)         // 4MHz
#define HZ2TMR1(x)      -(TMR1CLK/(x))

#define ERR_SENSOR0     1       // Sensor 0 error
#define ERR_SENSOR1     2       // Sensor 1 error
#define ERR_VCCLO       3       // VCC low (<4.5V, operation continue)
#define ERR_VCCCRIT     4       // VCC critical (<4.0V)
#define ERR_THERMAL     5       // Thermal protection
#define ERR_SENSDIFF    6       // Sensor difference too large

#define F_VCOMP         0x01    // VCC compensation

/******************************************************************************/
/* EEPROM Default                                                             */
/******************************************************************************/
#if sizeof(float) == 3
#error "Use 32bit float memory model"
#else
// 32bit float memory model (t=28, kp=1, ki=0.001, kd=0, sens=0, interval=1)
__EEPROM_DATA(0x00,0x00,0xE0,0x41,0x00,0x00,0x80,0x3F);
__EEPROM_DATA(0x6E,0x12,0x83,0x3A,0x00,0x00,0x00,0x00);
__EEPROM_DATA(0x00,0x01,0x00,0x00,0x00,0x00,0x00,0x00);
#endif

/******************************************************************************/
/* Main                                                                       */
/******************************************************************************/
void main()
{
    float t1, t2, tavg, tsrc, vcc, duty, vratio;
    unsigned char dh_vcclo = 0;

#if(USB_SPEED_OPTION == USB_FULL_SPEED)
    OSCTUNE = 0x80; //3X PLL ratio mode selected
    OSCCON = 0x70;  //Switch to 16MHz HFINTOSC
    OSCCON2 = 0x10; //Enable PLL, SOSC, PRI OSC drivers turned off
    while(OSCCON2bits.PLLRDY != 1); //Wait for PLL lock
    ACTCON = 0x90;  //Enable active clock tuning for USB operation
#endif

    // Setup port direction
    TRISA = 0x00;               // All output (7seg led)
    ANSELA = 0x00;              // All digital
    TRISB = 0x03;               // RB0 as SDA(I/O) set to 1
    ANSELB = 0x00;              // All digital
    TRISC = 0xC0;               // RC7 as AN19, RC6 in
    ANSELC = 0x80;              // AN19 VCC measure
    TRISD = 0x00;               // Port D not used
    ANSELD= 0x00;               // All digital
    TRISE = 0x00;               // All (except RE3/MCLR) output
    ANSELE = 0x00;

    // Lamp check (All segments lit up to 2s)
    LATA = 0xFF;
    LATC = 0x03;
    LATE = 0x07;
    for(tick=0; tick<200; tick++){ __delaywdt_ms(10);}

    // Setup Timer 0 (100ms periodic interrupt)
    T0CON = 0b10000011;
    INTCONbits.TMR0IE = 1;  // Enable
    INTCON2bits.TMR0IP = 0; // Low priority

    // Setup Timer 1 (Dynamic display 4kHz interrupt)
    T1CON = 0b00000111;
    PIE1bits.TMR1IE = 1;    // Enable
    IPR1bits.TMR1IP = 0;    // Low priority

    // Setup Timer 2 (for PWM)
    T2CON = 0b00000110;
    PR2 = 255;
    // Setup PWM
    CCP1CON = 0b00001100;   // PWM mode

    // Setup MSSP (I2C master mode, interrupt not used)
    SSP1STATbits.SMP = 1;
    SSP1CON1 = 0x28;        // I2C Master mode, clk = FOSC / (4 * (SSP1ADD +1))
    SSP1ADD = 39;           // clk = 100kHz

    // Setup A/D
    ADCON0 = 0b01001101;    // AN19, ADON
    ADCON1 = 0b00001000;    // Vref+ = FVR, Vref-=AVSS
    ADCON2 = 0b10000111;    // RC Clock
    VREFCON0 = 0b10100000;  // FVR ON, 2x (Ref=2.048V)

    // Enable interrupt
    RCONbits.IPEN = 1;      // Priority enable
    INTCONbits.PEIE = 1;
    INTCONbits.GIEH = 1;
    INTCONbits.GIEL = 1;

    // Setup USB (USB interrupt enabled inside following function)
    USBDeviceInit();
    USBDeviceAttach();

    // Init temp sensor
    temp_init(0);
    temp_init(1);

    // Init state buffer
    tfbuf_init(temp_read(0), &tfbuf[0]);
    tfbuf_init(temp_read(1), &tfbuf[1]);
    tfbuf_init(0.0F, &tfbuf[2]);
    tfbuf_init(0.0F, &tfbuf[3]);

    // Read configuration
    conf_read();

    tick = 0;
    // Task loop
    while(1){
#ifdef USB_POLLING
        USBDeviceTasks();
#endif
        if(USBGetDeviceState() == CONFIGURED_STATE){
            conf_dialog();
            CDCTxService();
        }

        if(xtsk_100ms){
            // 100ms periodic task
            xtsk_100ms = 0;

            // Measure supply voltage
            vcc = adc_getvcc();

            // Measure temperature
            t1 = temp_read(0);
            t2 = temp_read(1);

            // Apply 0.1Hz LPF
            t1 = lpf(t1, &tfbuf[0]);
            t2 = lpf(t2, &tfbuf[1]);

            // Sensor difference error detection
            if((t1>t2 ? t1-t2 : t2-t1) > 2.0F){
                disp_err(ERR_SENSDIFF);
                main_halt();
            }
            
            // Average
            tavg = (t1 + t2) /2.0F;

            // Select sensor
            switch(conf.st.src){
                default:
                case 0:
                    tsrc = tavg;
                    break;
                case 1:
                    tsrc = t1;
                    break;
                case 2:
                    tsrc = t2;
                    break;
            }

            // PID-controller
            duty = pid_control(
                    conf.st.tset,       // Set tepmerature
                    tsrc,               // Current temperature
                    conf.st.pgain,      // P gain
                    conf.st.igain,      // I gain
                    conf.st.dgain,      // D gain
                    &tfbuf[2],          // I state buffer
                    &tfbuf[3]           // D state buffer
                    );

            // VCC compensation
            if(conf.st.flags & F_VCOMP){
                if(vcc < 1.0F) vcc = 1.0F;  // Avoid zero div
                vratio = 5.0F /vcc;
                duty = duty * vratio * vratio;
            }

            // PWM control
            pwm_setduty(duty);

            // Display
            if(vcc < 4.5F){
                dh_vcclo = 10;
            }
            if(dh_vcclo > 0){
                // Low voltage alarm
                if(tick % 20 < 10){
                    disp_err(ERR_VCCLO);
                } else {
                    disp_temp(tsrc);
                }
                dh_vcclo --;
            } else {
                // Normal display
                disp_temp(tsrc);
                //disp_debug();
                //disp_ul(count);
            }

            // Talker
            if(USBGetDeviceState() == CONFIGURED_STATE){
                talk(tick, t1, t2, tavg, duty, vcc);
            }

            // Low voltage protection
            if(vcc < 4.0F){
                disp_err(ERR_VCCCRIT);
                main_halt();
            }

            // Thermal protection
            if(PORTCbits.RC6){
                disp_err(ERR_THERMAL);
                main_halt();
            }

            tick++;
            CLRWDT();
        }
    }
}

// Stop control
void main_halt()
{
    pwm_setduty(0.0F);
    putrsUSBUSART("Halt\r\n");
    while(1) {
#ifdef USB_INTERRUPT
        USBDeviceAttach();
#endif
#ifdef USB_POLLING
        USBDeviceTasks();
#endif
        CDCTxService();
        CLRWDT();
    }
}

/******************************************************************************/
/* Interrupt handler                                                          */
/******************************************************************************/
// Low priority interrupt
void interrupt low_priority isr_low()
{
    // Timer 0 for 100ms periodic task
    if(INTCONbits.TMR0IF){
        INTCONbits.TMR0IF = 0;
        // 100ms periodic timer interrupt
        TMR0 = MSEC2TMR0(100);
        xtsk_100ms = 1;
    }

    // Timer 1 for dynamic display scan
    if(PIR1bits.TMR1IF){
        PIR1bits.TMR1IF = 0;
        TMR1 = HZ2TMR1(4000);
        if(disp_seq < 79){
            disp_seq++;
        } else {
            disp_seq = 0;
        }
        switch(disp_seq & 0x0f){
            case 0:
                led_blank();
                break;
            case 1:
                led_digit((disp_seq & 0xf0) >> 4);
                break;
            default:
                break;
        }
    }
}

// High priority interrupt
void interrupt high_priority isr_high()
{
#ifdef USB_INTERRUPT
    // Flag cleared inside following function
    USBDeviceTasks();
#endif
}

/******************************************************************************/
/* Talker                                                                     */
/******************************************************************************/
void talk(unsigned long tick, float t1, float t2, float tavg, float duty, float vcc)
{
    char buff[64], temp_buff[16], *p;
    int len;

    if(dlg != DLG_NONE) return;                 // Stop talking when dialog is open
    if(conf.st.tint == 0) return;               // Talker disabled
    if(tick % (conf.st.tint *10) != 0) return;  // Talk at specified interval

    p = buff;

    len = sprintf(temp_buff, "%ld,", tick /10);
    strcpy(p, temp_buff);
    p += len;

    len = sprintf(temp_buff, "%f,", t1);
    strcpy(p, temp_buff);
    p += len;

    len = sprintf(temp_buff, "%f,", t2);
    strcpy(p, temp_buff);
    p += len;

    len = sprintf(temp_buff, "%f,", tavg);
    strcpy(p, temp_buff);
    p += len;

    len = sprintf(temp_buff, "%.3f,", duty);
    strcpy(p, temp_buff);
    p += len;

    len = sprintf(temp_buff, "%.2f\r\n", vcc);
    strcpy(p, temp_buff);
    p += len;

    if(USBUSARTIsTxTrfReady()){
        putsUSBUSART(buff);
    }
}

/******************************************************************************/
/* Parameter configuration                                                    */
/******************************************************************************/
// Read configuration from EEPROM
void conf_read()
{
    char n;
    for(n=0; n<sizeof(conf); n++){
        conf.data[n] = eep_read(n);
    }
}

// Write configuration to EEPROM
void conf_write()
{
    char n;
    for(n=0; n<sizeof(conf); n++){
        eep_write(n, conf.data[n]);
    }
}

// Show configuration
void conf_show()
{	
    char buff[255], temp_buff[32];
    char *p;
    int len;
    p = buff;

    strcpy(p, "<Show>\r\n");
    p += 8;

    len = sprintf(temp_buff, " Tset:%f\r\n", conf.st.tset);
    strcpy(p, temp_buff);
    p += len;

    len = sprintf(temp_buff, " Pgain:%f\r\n", conf.st.pgain);
    strcpy(p, temp_buff);
    p += len;

    len = sprintf(temp_buff, " Igain:%f\r\n", conf.st.igain);
    strcpy(p, temp_buff);
    p += len;

    len = sprintf(temp_buff, " Dgain:%f\r\n", conf.st.dgain);
    strcpy(p, temp_buff);
    p += len;

    len = sprintf(temp_buff, " Source:%d\r\n", conf.st.src);
    strcpy(p, temp_buff);
    p += len;

    len = sprintf(temp_buff, " Interval:%d\r\n", conf.st.tint);
    strcpy(p, temp_buff);
    p += len;

    len = sprintf(temp_buff, " Vcomp:%s\r\n", conf.st.flags & F_VCOMP ? "on" : "off");
    strcpy(p, temp_buff);
    p += len;

    if(USBUSARTIsTxTrfReady()){
        putsUSBUSART(buff);
    }
}

float temp_value;
unsigned char after_decimal;

// Clear dialog value
#define value_clear() do{temp_value = 0.0F; after_decimal = 0;} while(0)

// Retrieve dialog value
#define value_get() temp_value
#define value_get_uc()  (unsigned char)((temp_value < 0.0F) ? 0 : ( (temp_value > 255.0F) ? 255 : temp_value +0.5))

// Parse dialog input
void value_parse(char c)
{
    float temp;
    unsigned char n;

    if(c == '.'){
        if(after_decimal == 0){
            after_decimal = 1;
        }
        return;
    }
    if(after_decimal == 0){
        temp_value *= 10.0F;
        temp_value += (float)(c - '0');
    } else {
        temp = (float)(c - '0');
        for(n=0; n < after_decimal; n++){
            temp = temp/10.0F;
        }        
        temp_value += temp;
        after_decimal++;
    }
}

// Dialog menu
#define MENU_MSG \
"<Config>\r\n\
 t:Tset\r\n\
 p:Pgain\r\n\
 i:Igain\r\n\
 d:Dgain\r\n\
 o:Source(0=avg/1=sensor1/2=sensor2)\r\n\
 k:Talker interval(0-255)\r\n\
 v:VCC comp\r\n\
 s:Show\r\n\
 w:Write EEPROM\r\n\
 r:Read EEPROM\r\n\
 z:Clear tf buff\r\n\
 f:FW Ver\r\n\
 u:FW Update(99=confirm)\r\n\
 ?:Menu\r\n\
 x:Exit\r\n"

// Configuration dialog
void conf_dialog()
{
    unsigned char buff[8], c, tmp;
    int len, n;
    len = getsUSBUSART(buff, sizeof(buff));
    for(n=0; n<len; n++){
        c = buff[n];
        switch(dlg){
            case DLG_NONE:
                if(c == 'c'){   // 'c' -> Enter dialog
                    dlg = DLG_MENU;
                    putrsUSBUSART(MENU_MSG);
                }
                break;
            case DLG_MENU:
                switch(c){
                    case 't':   // 't' -> Temperature set
                        putrsUSBUSART("Tset:");
                        dlg = DLG_TSET;
                        break;
                    case 'p':   // 'p' -> P gain set
                        putrsUSBUSART("Pgain:");
                        dlg = DLG_PGAIN;
                        break;
                    case 'i':   // 'i' -> I gain set
                        putrsUSBUSART("Igain:");
                        dlg = DLG_IGAIN;
                        break;
                    case 'd':   // 'd' -> D gain set
                        putrsUSBUSART("Dgain:");
                        dlg = DLG_DGAIN;
                        break;
                    case 's':   // 's' -> Show current
                        conf_show();
                        break;
                    case 'w':   // 'w' -> Write EEPROM
                        conf_write();
                        putrsUSBUSART("<Write>\r\n");
                        break;
                    case 'r':   // 'r' -> Read EEPROM
                        conf_read();
                        putrsUSBUSART("<Read>\r\n");
                        break;
                    case 'z':   // 'z' -> Clear state buffer
                        tfbuf_init(0.0F, &tfbuf[2]);
                        tfbuf_init(0.0F, &tfbuf[3]);
                        putrsUSBUSART("<Clear>\r\n");
                        break;
                    case 'o':   // 'o' -> Source
                        putrsUSBUSART("Source:");
                        dlg = DLG_SRC;
                        break;
                    case 'k':   // 'k' -> Talker interval
                        putrsUSBUSART("Interval:");
                        dlg = DLG_TINT;
                        break;
                    case 'v':   // 'v' -> VCC comp
                        if(conf.st.flags & F_VCOMP){
                            conf.st.flags &= ~F_VCOMP;
                            putrsUSBUSART("<Vcomp off>\r\n");
                        } else {
                            conf.st.flags |= F_VCOMP;
                            putrsUSBUSART("<Vcomp on>\r\n");
                        }
                        break;
                    case 'x':   // 'x' -> Exit dialog
                        putrsUSBUSART("<Exit>\r\n");
                        dlg = DLG_NONE;
                        break;
                    case '?':   // '?' -> Menu
                        putrsUSBUSART(MENU_MSG);
                        break;
                    case  'f':  // 'f' -> FW Ver
                        putrsUSBUSART("<FW Ver>\r\n $Rev: 25 $\r\n");
                        break;
                    case 'u':   // 'u' -> FW Update
                        putrsUSBUSART("Confirm:");
                        dlg = DLG_FWUP;
                        break;
                    default:
                        break;
                }

                value_clear();
                break;

            case DLG_TSET:
            case DLG_PGAIN:
            case DLG_IGAIN:
            case DLG_DGAIN:
            case DLG_SRC:
            case DLG_TINT:
            case DLG_FWUP:
                if(c == '\r'){  // '\r' -> End input
                    switch(dlg){
                        case DLG_TSET:
                            conf.st.tset = value_get();
                            if(conf.st.tset > 35.0F){
                                conf.st.tset = 35.0F;
                            }
                            break;
                        case DLG_PGAIN:
                            conf.st.pgain = value_get();
                            break;
                        case DLG_IGAIN:
                            conf.st.igain = value_get();
                            break;
                        case DLG_DGAIN:
                            conf.st.dgain = value_get();
                            break;
                        case DLG_SRC:
                            if(value_get_uc() > 2){
                                conf.st.src = 2;
                            } else {
                                conf.st.src = value_get_uc();
                            }
                            break;
                        case DLG_TINT:
                            conf.st.tint = value_get_uc();
                            break;
                        case DLG_FWUP:
                            if(value_get_uc() == 99){
                                invoke_fwloader();
                            }
                            break;
                        default:
                            break;
                    }
                    putrsUSBUSART("\r\n");
                    dlg = DLG_MENU;
                } else if(c == '.' || isdigit(c)){  // valid letter
                    value_parse(c);
                    buff[0] = c;    // Echo back
                    buff[1] = 0;
                    putsUSBUSART(buff);
                } else {    // Invalid letter
                    putrsUSBUSART("ABORT!!\r\n");
                    dlg = DLG_MENU;
                }
                break;
            default:
                break;
        }
    }
}

/******************************************************************************/
/* LPF, PID-controller                                                        */
/******************************************************************************/
#define TS      (0.1F) // Main tick (100ms)

// LPF (1st order)
//     wc
//  --------
//   S + wc
#define LPF_WC (2.0F*3.14F*0.1F)  // Cutoff 0.1Hz
#define LPF_N0 ((float)(LPF_WC*TS/(2.0F+LPF_WC*TS)))            // num[0]
#define LPF_N1 ((float)(LPF_WC*TS/(2.0F+LPF_WC*TS)))            // num[1]
#define LPF_D1 ((float)((-2.0F+LPF_WC*TS)/(2.0F+LPF_WC*TS)))    // den[1]

// Psuedo-integral (1/s ~= 1/(s+a) where a << 1)
//      1
//  ----------
//   S + INT_A
#define INT_A   (0.0000F)
#define INT_N0  ((float)(TS/(2.0F+INT_A*TS)))                   // num[0]
#define INT_N1  ((float)(TS/(2.0F+INT_A*TS)))                   // num[1]
#define INT_D1  ((float)((-2.0F+INT_A*TS)/(2.0F+INT_A*TS)))     // den[1]

// Psuedo-derivative (s ~= as/(s+a) where a >> 1)
//    S*DIF_A
//  -----------
//   S + DIF_A
#define DIF_A   (10000.0F)
#define DIF_N0  ((float)(( 2.0F*DIF_A)/(2.0F+DIF_A*TS)))        // num[0]
#define DIF_N1  ((float)((-2.0F*DIF_A)/(2.0F+DIF_A*TS)))        // num[1]
#define DIF_D1  ((float)((-2.0F+DIF_A*TS)/(2.0F+DIF_A*TS)))     // den[1]

// Initialize both input/output buffer at given value
void tfbuf_init(float in, struct tfbuf_s *pbuf)
{
    pbuf->in_z1 = in;
    pbuf->out_z1 = in;
}

// Low pass filter
float lpf(float in, struct tfbuf_s *pbuf)
{
    float tmp;

    tmp  = in * LPF_N0;
    tmp += pbuf->in_z1 * LPF_N1;
    tmp -= pbuf->out_z1 * LPF_D1;

    pbuf->in_z1 = in;
    pbuf->out_z1 = tmp;

    return tmp;
}

// PID controller
float pid_control(float obj, float curr, float pgain, float igain, float dgain, struct tfbuf_s *ibuf, struct tfbuf_s *dbuf)
{
    float err, pout, iout, dout;

    // Error
    err = obj - curr;

    // P Control
    pout = err * pgain;

    // I Control
    if(pout < 1.0F){
        iout  = err * INT_N0;
        iout += ibuf->in_z1 * INT_N1;
        iout -= ibuf->out_z1 * INT_D1;
        ibuf->in_z1 = err;
        ibuf->out_z1 = iout;
        iout = iout * igain;
    } else {
        tfbuf_init(0.0F, ibuf);
        iout = 0.0F;
    }

    // D Control
    dout  = err * DIF_N0;
    dout += dbuf->in_z1 * DIF_N1;
    dout -= dbuf->out_z1 * DIF_D1;
    dbuf->in_z1 = err;
    dbuf->out_z1 = dout;
    dout = dout * dgain;

    // Return P+I+D
    return pout + iout + dout;
}

/******************************************************************************/
/* Temperature sensor                                                         */
/******************************************************************************/
const unsigned char i2c_addr[2] = {0x48, 0x49};
#define ADDR_READ(x)    (((x) << 1) | 0x01)
#define ADDR_WRITE(x)   ((x) << 1)
//#define REGDATA_CONFIG 0b11000000    // 1sps, 16bit resolution
#define REGDATA_CONFIG 0b10000000      // continuous, 16bit resolution
#define REGADDR_CONFIG 0x03
#define REGADDR_TEMP   0x00
void temp_init(unsigned char ch)
{
    unsigned char data;
    // Write configuration register
    i2c_start();
    i2c_send(ADDR_WRITE(i2c_addr[ch]));
    i2c_send(REGADDR_CONFIG); // Set pointer
    i2c_send(REGDATA_CONFIG); // Write data
    i2c_stop();

    // Readback configuration register
    i2c_start();
    i2c_send(ADDR_WRITE(i2c_addr[ch]));
    i2c_send(REGADDR_CONFIG); // Set pointer
    i2c_restart();
    i2c_send(ADDR_READ(i2c_addr[ch]));
    data = i2c_recv();
    i2c_sendnack();
    i2c_stop();

    // Check coincidence
    if(data == REGDATA_CONFIG) return;  // OK

    // Error
    disp_err(ERR_SENSOR0 + ch);
    main_halt();
}

float temp_read(unsigned char ch)
{
    short data;
    static short prev_data[2];
    static unsigned char err_count[2];

    // Set register address
    i2c_start();
    i2c_send(ADDR_WRITE(i2c_addr[ch]));
    if(!i2c_isacked()) goto temp_error;
    i2c_send(REGADDR_TEMP);
    if(!i2c_isacked()) goto temp_error;

    // Read temperature
    i2c_restart();
    i2c_send(ADDR_READ(i2c_addr[ch]));
    if(!i2c_isacked()) goto temp_error;
    data = (short)(i2c_recv()) << 8;
    i2c_sendack();
    // Reg addr increment automatically
    data |= i2c_recv();
    i2c_sendnack();
    i2c_stop();

    // Success
    prev_data[ch] = data;   // Save current result
    err_count[ch] = 0;      // Clear error counter
    return ((float)data)/128.0F;

    // Error
temp_error:
    err_count[ch] ++;
    if(err_count[ch] > 2){
        disp_err(ERR_SENSOR0 + ch);
        main_halt();
    }
    return ((float)prev_data[ch])/128.0F;   // Use previous result
}

/******************************************************************************/
/* PWM duty control                                                           */
/******************************************************************************/
#define PWM_MINDUTY  (0.0F)
#define PWM_MAXDUTY  (1.0F)
#define PWM_MAXWIDTH (1023)
void pwm_setduty(float duty)
{
    unsigned short width;

    if(duty < PWM_MINDUTY) duty = PWM_MINDUTY;
    if(duty > PWM_MAXDUTY) duty = PWM_MAXDUTY;
    width = (unsigned short)(duty * (float)PWM_MAXWIDTH);

    // Output
    CCPR1L = (unsigned char) (width >> 2);              // Upper 8bit
    CCP1CONbits.DC1B = (unsigned char)(width & 0x03);   // Lower 2bit
}

/******************************************************************************/
/* A/D conversion (VCC measurement)                                           */
/******************************************************************************/
#define R1      (10.0F)     // Divider Hi-side 10kohm
#define R2      (5.1F)      // Divider Lo-side 5.1kohm
#define VREF    (2.048F)    // FVR 1.024 x 2
#define VLSB    ((float)(VREF/1024.0F*(R1+R2)/R2))
float adc_getvcc(void)
{
    short adraw;
    // __delay_us(5);              // Acquision time (>4.42us)

    ADCON0bits.GO_nDONE = 1;    // Start conversion
    while(ADCON0bits.GO_nDONE) continue;    // Wait until done

    adraw = (((short)(ADRESH & 0x03)) << 8) | (short)ADRESL;

    return (float)adraw * VLSB;
}

/******************************************************************************/
/* LED Display                                                                */
/******************************************************************************/
// Port A font table
//                    A F G D.
#define SEGMASK_A   0b10101011
#define SEGBIT_ADP  0b00000001
const unsigned char segbit_A[] = {
    //A F G D.
    0b10100010, // 0
    0b00000000, // 1
    0b10001010, // 2
    0b10001010, // 3
    0b00101000, // 4
    0b10101010, // 5
    0b10101010, // 6
    0b10000000, // 7
    0b10101010, // 8
    0b10101010, // 9
    0b10101000, // A
    0b00101010, // b
    0b10100010, // C
    0b00001010, // d
    0b10101010, // E
    0b10101000, // F
    0b00001000, // -
    0b00000000  // BLANK
};

// Port C font table
//                          CE
#define SEGMASK_C   0b00000011
const unsigned char segbit_C[] = {
    //      CE
    0b00000011, // 0
    0b00000010, // 1
    0b00000001, // 2
    0b00000010, // 3
    0b00000010, // 4
    0b00000010, // 5
    0b00000011, // 6
    0b00000010, // 7
    0b00000011, // 8
    0b00000010, // 9
    0b00000011, // A
    0b00000011, // b
    0b00000001, // C
    0b00000011, // d
    0b00000001, // E
    0b00000001, // F
    0b00000000, // -
    0b00000000  // BLANK
};

// Port E font table
//                          B
#define SEGMASK_E   0b00000010
const unsigned char segbit_E[] = {
    //      B
    0b00000010, // 0
    0b00000010, // 1
    0b00000010, // 2
    0b00000010, // 3
    0b00000010, // 4
    0b00000000, // 5
    0b00000000, // 6
    0b00000010, // 7
    0b00000010, // 8
    0b00000010, // 9
    0b00000010, // A
    0b00000000, // b
    0b00000000, // C
    0b00000010, // d
    0b00000000, // E
    0b00000000, // F
    0b00000000, // -
    0b00000000  // BLANK
};

// Drive segment
void led_seg(unsigned char c)
{
    if(c & SEG_DP){
        c &= ~SEG_DP;
        LATA = LATA & ~SEGMASK_A | segbit_A[c] | SEGBIT_ADP;
    } else {
        LATA = LATA & ~SEGMASK_A | segbit_A[c];
    }
    LATC = LATC & ~SEGMASK_C | segbit_C[c];
    LATE = LATE & ~SEGMASK_E | segbit_E[c];
}

// All digit off
void led_blank(void)
{
    LATAbits.LA6 = 0;   // LEFT (10000's)
    LATEbits.LE2 = 0;
    LATEbits.LE0 = 0;
    LATAbits.LA4 = 0;
    LATAbits.LA2 = 0;   // RIGHT (1's)
}

unsigned char led_buff[5];
// Lit up specified digit
void led_digit(unsigned char digit)
{
    led_seg(led_buff[digit]);
    switch(digit){
        case 0: // LEFT (10000's)
            LATAbits.LA6 = 1;
            break;
        case 1:
            LATEbits.LE2 = 1;
            break;
        case 2:
            LATEbits.LE0 = 1;
            break;
        case 3:
            LATAbits.LA4 = 1;
            break;
        case 4: // RIGHT (1's)
            LATAbits.LA2 = 1;
            break;
        default:
            break;
    }
}

// Debug
void disp_debug()
{
    unsigned char reg1, reg2;
    
    reg1 = ACTCON;
    reg2 = USBDeviceState;
    
    led_buff[4] = OSCTUNE & 0x0f;
    led_buff[3] = reg2 & 0x0f;
    led_buff[2] = (reg2 >> 4) & 0x0f;
    led_buff[1] = reg1 & 0x0f;
    led_buff[0] = (reg1 >> 4) & 0x0f;
}

// Display long value
void disp_ul(unsigned long ul)
{
    led_buff[4] = ul % 10;
    led_buff[3] = (ul % 100)/10;
    led_buff[2] = (ul % 1000)/100;
    led_buff[1] = (ul % 10000)/1000;
    led_buff[0] = (ul % 100000)/10000;
}

// Display temperature
void disp_temp(float temp)
{
    short u;
    // Rounding
    if(temp > 0){
        u = (short)(temp * 100.0F + 0.5F);
    } else {
        u = (short)-(temp * 100.0F - 0.5F);
    }
    led_buff[4] = 0x0C;                         // 'C' for celsius
    if(tick % 10 < 5) led_buff[4] |= SEG_DP;    // Blinking DP
    led_buff[3] = u % 10;                       // 0.01's
    led_buff[2] = (u % 100)/10;                 // 0.1's
    led_buff[1] = ((u % 1000)/100) | SEG_DP;    // 1's
    led_buff[0] = (temp < 0) ? SEG_DASH : ((u % 10000)/1000);   // 10's
}

// Display error code
void disp_err(unsigned char err)
{
    led_buff[4] = err % 10;           // 1's
    led_buff[3] = (err % 100)/10;     // 10's
    led_buff[2] = (err % 1000)/100;   // 100's
    led_buff[1] = SEG_DASH;           // '-'
    led_buff[0] = 0x0E;               // 'E'
}

/******************************************************************************/
/* EEPROM                                                                     */
/******************************************************************************/
// Read data from eeprom
unsigned char eep_read(unsigned char addr)
{
    EEADR = addr;
    EECON1bits.CFGS = 0;
    EECON1bits.EEPGD =0;
    EECON1bits.RD = 1;
    return EEDATA;
}

// Write data to eeprom
void eep_write(unsigned char addr, unsigned char data)
{
    EEADR = addr;
    EEDATA = data;
    EECON1bits.CFGS = 0;
    EECON1bits.EEPGD = 0;
    EECON1bits.WREN = 1;
    INTCONbits.GIE = 0;
    EECON2 = 0x55;
    EECON2 = 0xaa;
    EECON1bits.WR = 1;
    INTCONbits.GIE =1;
    EECON1bits.WREN = 0;
    while(EECON1bits.WR) continue;
}

/******************************************************************************/
/* I2C                                                                        */
/******************************************************************************/
#if 0 // Not used (init at main)
void i2c_init(void)
{
    // Setup MSSP (I2C master mode, interrupt not used)
    SSP1STATbits.SMP = 1;
    SSP1CON1 = 0x28;        // I2C Master mode, clk = FOSC / (4 * (SSP1ADD +1))
    SSP1ADD = 39;           // clk = 100kHz
}
#endif

#if 0 // Not used (use macro function instead)
void i2c_start()
{
    SSP1CON2bits.SEN = 1;
    while(SSP1CON2bits.SEN) continue;
}
#endif

#if 0 // Not used (use macro function instead)
void i2c_restart()
{
    PIR1bits.SSPIF = 0;
    SSP1CON2bits.RSEN = 1;
    while (SSP1CON2bits.RSEN) continue;
}
#endif

#if 0 // Not used (use macro function instead)
void i2c_stop()
{
    PIR1bits.SSPIF = 0;
    SSP1CON2bits.PEN = 1;
    while (SSP1CON2bits.PEN) continue;
    PIR1bits.SSPIF = 0;
}
#endif

#if 0 // Not used (use macro function instead)
void i2c_send(unsigned char c)
{
    PIR1bits.SSPIF = 0;
    SSP1BUF = c;
    while (!PIR1bits.SSPIF) continue;
}
#endif

unsigned char i2c_recv()
{
    SSP1CON2bits.RCEN = 1;
    while (SSP1CON2bits.RCEN) continue;
    return(SSP1BUF);
}

#if 0 // Not used (use macro function instead)
void i2c_sendack()
{
    SSP1CON2bits.ACKDT = 0;
    SSP1CON2bits.ACKEN = 1;
    while (SSP1CON2bits.ACKEN) continue;
}
#endif

#if 0 // Not used (use macro function instead)
void i2c_sendnack()
{
    SSP1CON2bits.ACKDT = 1;
    SSP1CON2bits.ACKEN = 1;
    while (SSP1CON2bits.ACKEN) continue;
}
#endif

// FW Loader invocation
void invoke_fwloader(void)
{
    unsigned char c;
    pwm_setduty(0.0F);  // Stop heater
    INTCON = 0;         // Disable interrupt
    ADCON0 = 0;         // Disable A/D
    LATA = 0;           // LED OFF
    LATC = 0;           // LED OFF
    LATE = 0;           // LED OFF
    USBDeviceDetach();  // Cutoff USB
    for(c=0; c<10; c++) __delay_ms(10);
    asm("goto 0x30");   // Jump to FW loader
}
// EOF